/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>

#include "Defines.h"
#include "Gouraud.h"

extern unsigned char *_w_screen;

int YADDtable[MAXY];

int iLeftEdge, iRightEdge, iLeftEdgeColor;

void initgour()
{
    int i;

    for (i = 0; i < MAXY; i++)
	YADDtable[i] = MAXX * i;
}

/*
 Gouraud Triangle Filler
*/

void GhLine(unsigned char* m, int lle, int rre, int hsc)
{
    int hcd;
    unsigned char *DP;
    unsigned int len, liLeftEdgeColor;

    liLeftEdgeColor = iLeftEdgeColor;
    if (lle > MAXX - 1)
	return;
    if (rre < 0)
	return;
    if (lle < 0) {
	liLeftEdgeColor += (-lle) * hsc;
	lle = 0;
    }
    if (rre > MAXX - 1)
	rre = MAXX - 1;
    if ((lle == rre) || (lle == rre - 1) || (lle>rre)) {
	hcd = liLeftEdgeColor;
	*(m + lle) = (BYTE) (hcd >> 16);
	return;
    }
    DP = m + lle;
    len = (rre - lle);
    hcd = liLeftEdgeColor;
    while (len--) {
    	if(!(DP>=_w_screen && DP<_w_screen+MAXX*MAXY))
	{
	    __asm __volatile ("int $3");
	}
	*DP++ = (BYTE) (hcd >> 16);
	hcd += hsc;
    }
    hcd -= hsc;
    *DP = (BYTE) (hcd >> 16);
}

void GouraudFill_Tri(LPOLY * data)
{
    LPOINT *p1, *p2, *p3;
    LPOINT *f, *m, *l;
    int dxFL, dxFM, dxML;
    int dyFL, dyFM, dyML;
    int sFL, sFM, sML, sl, sr;
    int iSlopeColorLeftEdge, iSlopeColorRightEdge;
    int hsc;

    DWORD di;
    int fbot = FALSE, mleft;
    int fy, my;
    int cF, cM, cL;
    int cd;
    int off, off2, fx;

    p1 = data->P1;
    p2 = data->P2;
    p3 = data->P3;

    //sort points
    if (p1->Y < p2->Y) {
	if (p1->Y < p3->Y) {
	    if (p2->Y < p3->Y) {
		f = p1;
		m = p2;
		l = p3;
	    }			//123
	    else {
		f = p1;
		m = p3;
		l = p2;
	    }			//132
	} else {
	    f = p3;
	    m = p1;
	    l = p2;
	}			//312 
    } else {
	if (p2->Y < p3->Y) {
	    if (p1->Y < p3->Y) {
		f = p2;
		m = p1;
		l = p3;
	    }			//213
	    else {
		f = p2;
		m = p3;
		l = p1;
	    }			//231
	} else {
	    f = p3;
	    m = p2;
	    l = p1;
	}			//321
    }

    //master y-clip
    if (f->Y >=MAXY || l->Y < 0)
	return;

    if (((f->X < 0) && (m->X < 0) && (l->X < 0)) ||
	((f->X >= MAXX) && (m->X >= MAXX) && (l->X >= MAXX)))
	return;

    if ((f->X > 32767) || (f->X < -32768) ||
	(m->X > 32767) || (m->X < -32768) ||
	(l->X > 32767) || (l->X < -32768)
	)
	return;

    //get Vertex color

    cF = f->Color;
    cM = m->Color;
    cL = l->Color;

    //calc slopes
    dxFL = l->X - f->X;		//line: First-Last
    if ((dyFL = l->Y - f->Y) != 0)
	sFL = (dxFL << 16) / dyFL;
    else
	return;

    dxML = l->X - m->X;		//line: Middle-Last
    if ((dyML = l->Y - m->Y) != 0)
	sML = (dxML << 16) / dyML;
    else
	fbot = TRUE;

    dxFM = m->X - f->X;		//line: First-Middle
    if ((dyFM = m->Y - f->Y) != 0)
	sFM = (dxFM << 16) / dyFM;
    else {
	fy = f->Y;		//=my
	if (f->X < m->X) {	//find left-right
	    sl = sFL;		//left slope + scansize adder
	    sr = sML;		//right slope + scansize adder
	    iSlopeColorLeftEdge = ((cL - cF) << 16) / dyFL;	//left color slope
	    if (dxFM == 0)
		hsc = 0;
	    else
		hsc = ((cM << 16) - (cF << 16)) / dxFM;	//horizontal color slope
	    if (fy>=MAXY)
	    	return;
	    else if (fy < 0) {
		off = 0 - fy;
		iLeftEdge = (f->X << 16) + (off * sFL);
		iRightEdge = (m->X << 16) + (off * sML);
		iLeftEdgeColor = (cF << 16) + (off * iSlopeColorLeftEdge);
		dyML -= off;
		di = (DWORD) _w_screen + YADDtable[0];	//di=base adx
	    } else {
		iLeftEdge = f->X << 16;	//left edge (le=int part)
		iRightEdge = m->X << 16;	//right edge (re=int part)
		iLeftEdgeColor = cF << 16;
		assert(fy<MAXY);
		di = (DWORD) _w_screen + YADDtable[fy];	//di=base adx
	    }
	} else {
	    sl = sML;
	    sr = sFL;
	    iSlopeColorLeftEdge = ((cL - cM) << 16) / dyML;	//left color slope
	    if (dxFM == 0)
		hsc = 0;
	    else
		hsc = ((cF<<16) - (cM<<16)) / (-dxFM);	//horizontal color slope
	    if (fy>=MAXY)
	    	return;
	    else if (fy < 0) {
		off = 0 - fy;
		iLeftEdge = (m->X << 16) + (off * sML);
		iRightEdge = (f->X << 16) + (off * sFL);
		iLeftEdgeColor = (cM << 16) + (off * iSlopeColorLeftEdge);
		dyML -= off;
		di = (DWORD) _w_screen + YADDtable[0];	//di=base adx
	    } else {
		iLeftEdge = m->X << 16;
		iRightEdge = f->X << 16;
		iLeftEdgeColor = cM << 16;
		assert(fy<MAXY);
		di = (DWORD) _w_screen + YADDtable[fy];	//di=base adx
	    }
	}
	goto flatTop;
    }

    // ScanConvert 1st part (flatBottom)
    if (sFM < sFL) {		//find left-right
	sl = sFM;		//middle left
	sr = sFL;
	iSlopeColorLeftEdge = ((cM - cF) << 16) / dyFM;
	iSlopeColorRightEdge = ((cL - cF) << 16) / dyFL;
	if ((cd = 2 + (f->X) + ((dyFM * sFL) >> 16) - (m->X)) == 0)
	    hsc = 0;
	else
	    hsc = ((cF << 16) + (dyFM * iSlopeColorRightEdge) - (cM << 16)) / cd;
	if (cd < 0)
	    return;		// th=-338, phi=-144, mi=840, springy1.plg
	mleft = TRUE;		//middle = left -> last = right (*)
    } else {
	sl = sFL;		//midle right (last left)
	sr = sFM;
	iSlopeColorLeftEdge = ((cL - cF) << 16) / dyFL;
	if ((cd = 2 + (m->X) - (f->X) - ((dyFM * sFL) >> 16)) == 0)
	    hsc = 0;
	else
	    hsc = (((cM - cF) << 16) - dyFM * iSlopeColorLeftEdge) / cd;
	if (cd < 0)
	    return;		// th=-338, phi=-144, mi=840, springy1.plg
	mleft = FALSE;
    }
    //y-clip
    fy = f->Y;
    my = m->Y;
    if (my < 0) {
	off = 0 - my;
	off2 = 0 - fy;
	if (mleft) {
	    sl = sML;
	    iSlopeColorLeftEdge = ((cL - cM) << 16) / dyML;	// Bug Fix!
	    iLeftEdge = (m->X << 16) + (off * sl);
	    iRightEdge = (f->X << 16) + (off2 * sr);
	    iLeftEdgeColor = (cM << 16) + (off * iSlopeColorLeftEdge);
	} else {
	    sr = sML;
	    iSlopeColorLeftEdge = ((cL - cF) << 16) / dyFL;
	    iLeftEdge = (f->X << 16) + (off2 * sl);
	    iRightEdge = (m->X << 16) + (off * sr);
	    iLeftEdgeColor = (cF << 16) + (off2 * iSlopeColorLeftEdge);
	}
	dyML -= off;
	di = (DWORD) _w_screen + YADDtable[0];	//di=base adx
	goto flatTop;
    }
    if (my >= MAXY) {
	dyFM -= (my - MAXY);
	fbot = TRUE;		//clipped -> fbot !
    }
    if (fy>=MAXY)
	return;
    else if (fy < 0) {
	off = 0 - fy;
	fx = (f->X << 16);
	iLeftEdge = fx + (off * sl);
	iRightEdge = fx + (off * sr);
	iLeftEdgeColor = (cF << 16) + (off * iSlopeColorLeftEdge);
	dyFM -= off;
	di = (DWORD) _w_screen + YADDtable[0];	//di=base adx
    } else {
	iLeftEdge = f->X << 16;	//left edge
	iRightEdge = iLeftEdge + 0x00007fff;	//=right edge (we add .5 though)
	iLeftEdgeColor = cF << 16;		//left edge color
	assert(fy<MAXY);
	di = (DWORD) _w_screen + YADDtable[fy];	//di=base adx
    }
    while (dyFM--) {
	GhLine((unsigned char*) (di), iLeftEdge>>16, iRightEdge>>16, hsc);	//draw Gouraudhline
	iLeftEdge += sl;
	iRightEdge += sr;		//update X and Y (thnx to the ScanAdder)
	iLeftEdgeColor += iSlopeColorLeftEdge;		//update left color
	di += MAXX;
    }
    if (fbot)
	return;			//if it is flatBottom we end here

    // ScanConvert 2nd part (flatTop)
    if (mleft) {
	sl = sML;		//middle is left
	iSlopeColorLeftEdge = ((cL - cM) << 16) / dyML;	//so we need a new color slope
    } else
	sr = sML;
  flatTop:			//if it is flatTop we start here
    //y-clip
    if (l->Y >=MAXY)
	dyML -= (l->Y - MAXY);
    while (dyML--) {
	GhLine((unsigned char*) (di), iLeftEdge>>16, iRightEdge>>16, hsc);
	iLeftEdge += sl;
	iRightEdge += sr;
	iLeftEdgeColor += iSlopeColorLeftEdge;
	di += MAXX;
    }
}

// coded and (c)1995 by George (tmL) Moshovitis 
// Lefteris (Freeman) Stamatogiannakis                 [11/1/95 - 22/2/95]
// Debugged, added 2D clipping and generally fixed-up by
// Thanassis Tsiodras                       [4/3/95]
