/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <SDL/SDL.h>

#include "3DTypes.h"
#include "Vgalib.h"
#include "Camera.h"
#include "Gouraud.h"
#include "Wu.h"
#include "Defines.h"


int main(int argc, char *argv[])
{
    Object3D *obj = NULL;
    FLOAT dphi=0.5,dth=0.5, step=25;

    Initialize();   

    if (argc!=3) {
	cout << "Usage: " << argv[0] << " filename.asc mode(0/1/2/3/4/5/6)\n";
	exit(0);
    }

    try {
	obj = new Object3D(argv[1]);
    }
    catch(RendererException *e) {
	e->DebugInfo();
	return -1;
    }

    initgour();
    init256();

    char windowTitle[256];

    strcpy(windowTitle, "trender SDL - ");
    strcat(windowTitle, argv[1]);
    SDL_WM_SetCaption(windowTitle, "trender");
    SDL_EnableUNICODE(1);

    switch(atoi(argv[2])) {
    case 0: // Points
    case 1: // Antialiased lines
    case 2: // Distance-shaded lines
    case 5: // Gouraud shaded polygons
    case 6: // Gouraud shaded polygons
	pal();
	break;
    case 3: // Flat shaded polygons
    case 4: // Flat shaded polygons with bounding lines
	set256Palette(obj->GetPaletteManager()->GetPalette());
	break;
    }

    Light torch(100,100,100, 255, 100,100,100);
    FLOAT fLightAngle = 45;

    torch.SetPosition(
	15*obj->GetRadiusX()*sin(fLightAngle*3.14159*2/360),
	15*obj->GetRadiusY()*cos(fLightAngle*3.14159*2/360),
	0);

    FLOAT th,phi,ddd;
    FLOAT vx,vy,vz; // eye point
    FLOAT px,py,pz; // lookat point

    th=180.0; phi=90.0;

    ddd=obj->GetRadiusZ()*step/100;
    if (obj->GetRadiusX()*step/100>ddd)
        ddd=obj->GetRadiusX()*step/100;
    if (obj->GetRadiusY()*step/100>ddd)
        ddd=obj->GetRadiusY()*step/100;

    vx = 2*obj->GetRadiusX(); vy = vz = 0;
    px = py = pz = 0;

    Camera sony(vx, vy, vz, px, py, pz);

    LONG frames=0, st=readtime();
    int flag = 0;

    int isDown=0, isUp=0, isLeft=0, isRight=0, isA=0, isZ=0, isW=0;

    while(1) {
        px=vx+ddd*2*sin(M_2PI*phi/360)*cos(M_2PI*th/360);
        py=vy+ddd*2*sin(M_2PI*phi/360)*sin(M_2PI*th/360);
        pz=vz+ddd*2*cos(M_2PI*phi/360);

	sony.Set(vx,vy,vz, px,py,pz);

	torch.SetPosition(
	    15*obj->GetRadiusX()*sin(fLightAngle*3.14159*2/360),
	    15*obj->GetRadiusY()*cos(fLightAngle*3.14159*2/360),
	    0);

	clear_page();

	switch(atoi(argv[2])) {
	case 0:
	    obj->RenderDots(&sony);
	    break;
	case 1:
	    obj->RenderWireFrame(&sony);
	    break;
	case 2:
	    obj->RenderDepthWireFrame(&sony);
	    break;
	case 3:
	    obj->RenderFlatWire(&sony, FALSE);
	    break;
	case 4:
	    obj->RenderFlatWire(&sony, TRUE);
	    break;
	case 5:
	    obj->RenderFastGouraud(&sony, &torch);
	    break;
	case 6:
	    obj->RenderZGouraud(&sony, &torch);
	    break;
	}
	show_page();

	frames++;

	SDL_Event event;

	if(SDL_PollEvent(&event)) {

	    switch(event.type) {
	    case SDL_KEYDOWN:
	    	{
		    switch(event.key.keysym.sym) {
		    case SDLK_UP:
		    	isUp = 1;
			break;
		    case SDLK_DOWN:
		    	isDown = 1;
			break;
		    case SDLK_LEFT:
		    	isLeft = 1;
			break;
		    case SDLK_RIGHT:
		    	isRight = 1;
			break;
		    case SDLK_a:
		    	isA = 1;
			break;
		    case SDLK_z:
		    	isZ = 1;
			break;
		    case SDLK_w:
		    	isW = 1;
			break;
		    case SDLK_SPACE:
			do {
			    do {
				SDL_PollEvent(&event);
			    } while (event.type != SDL_KEYUP);
			} while (event.key.keysym.sym != SDLK_SPACE);
			*argv[2] = '0'+((1+atoi(argv[2])) % 7);
			switch(atoi(argv[2])) {
			case 0: // Points
			case 1: // Antialiased lines
			case 2: // Distance-shaded lines
			case 5: // Gouraud shaded polygons
			case 6: // Gouraud shaded polygons
			    pal();
			    break;
			case 3: // Flat shaded polygons
			case 4: // Flat shaded polygons with bounding lines
			    set256Palette(obj->GetPaletteManager()->GetPalette());
			    break;
			}
		    	break;
		    case SDLK_ESCAPE:
		    	flag = 1;
		    default:;
		    }
		}
		break;
	    case SDL_KEYUP:
	    	{
		    switch(event.key.keysym.sym) {
		    case SDLK_UP:
		    	isUp = 0;
			break;
		    case SDLK_DOWN:
		    	isDown = 0;
			break;
		    case SDLK_LEFT:
		    	isLeft = 0;
			break;
		    case SDLK_RIGHT:
		    	isRight = 0;
			break;
		    case SDLK_a:
		    	isA = 0;
			break;
		    case SDLK_z:
		    	isZ = 0;
			break;
		    case SDLK_w:
		    	isW = 0;
			break;
		    case SDLK_ESCAPE:
		    	flag = 1;
		    default:;
		    }
		}
		break;
	    case SDL_QUIT:
		flag = 1;
		break;
	    }
	}
	if (isLeft) th=th+dth;
	if (isRight) th=th-dth;
	if (isDown) phi=phi+dphi;
	if (isUp) phi=phi-dphi;
	if (isA) {
	    vz=vz+ddd/5*cos(M_2PI*phi/360);
	    vx=vx+ddd/5*sin(M_2PI*phi/360)*cos(M_2PI*th/360);
	    vy=vy+ddd/5*sin(M_2PI*phi/360)*sin(M_2PI*th/360);
	}
	if (isZ) {
	    vz=vz-ddd/5*cos(M_2PI*phi/360);
	    vx=vx-ddd/5*sin(M_2PI*phi/360)*cos(M_2PI*th/360);
	    vy=vy-ddd/5*sin(M_2PI*phi/360)*sin(M_2PI*th/360);
	}
	if (isW)
	    fLightAngle+=3;
	if (phi>179.0) phi=179.0;
	if (phi<00) phi=1.0;

	if(flag)
	    break;
    }
    LONG en=readtime();

    close256();

    cout << "Scene has " << obj->GetNoOfTriangles() << " triangles and ";
    cout << obj->GetNoOfVertexes() << " vertexes.\n\n";
    cout << "Rendering speed: ";
    cout << (int)(((FLOAT)frames*(obj->GetNoOfTriangles()))/((en-st)/1000.0));
    cout << " triangles/sec (";
    cout << frames/((en-st)/1000.0) << " frames/sec).\n";
/*********************/

    if (obj)
	delete obj;

}
