/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "Types.h"
#include "3DTypes.h"
#include "Vgalib.h"
#include "Camera.h"
#include "Gouraud.h"
#include "Defines.h"

//#define LOG

#define MAX_VERTEXES_OF_SUBOBJECTS 65536

//--- Search for keyword in file pointed by FP
#define   fsearch(FP,str)     do {					\
				fscanf(FP,"%s",tmpSTR);			\
				if(feof(FP)) return;                  \
		              } while(strncmp(tmpSTR,str,strlen(str))!=0);

#define   fmysearch(FP,str)     do {					\
				fscanf(FP,"%s%[^:]",tmpSTR);			\
				if(feof(FP)) return;                  \
		              } while(strncmp(tmpSTR,str,strlen(str))!=0);

int ftrisearch(FILE *FP,char *str)
{
  char tmpSTR[80];
  do {
    fscanf(FP,"%s",tmpSTR);	
    if(feof(FP)) return(-2);
  } while(strncmp(tmpSTR,str,strlen(str))!=0);
  return(0);
}

//--- Search for char in file pointed by FP
#define   fsearchar(FP,chr)    while(fgetc(FP) != chr);

//--- Get Quoted string from file pointed by F
 
#define	getqstring(FP,str) {        \
    fsearchar(FP,'"');	            \
    fscanf(FP,"%[^\"]",str);    \
} /* skip 1st " read till 2nd " */

void Object3D::load_object(char *szFname, FLOAT fScale)
{
    FILE *in;
#ifdef LOG
    FILE *fpout = fopen("log.pas", "w");
#endif
    LONG nv, np;
    char tmpSTR[80], material[80];
    LONG vertexNR,polyNR,pointA,pointB,pointC;
    FLOAT fVertexX, fVertexY, fVertexZ;
    BOOLEAN final_pass = FALSE;
    Vertex *SubObjectTable[MAX_VERTEXES_OF_SUBOBJECTS];
    char *fname = new char[strlen(szFname)+10];

    cout << "Preparing Object Loading...\n";
    strcpy(fname,szFname);
    if (!strrchr(fname,'.')) {
        strcat(fname,".asc");
    }

    while(1) {
        in = fopen(fname, "r");
        if (!in)
            throw new RendererException("Can't open file in Object construction");

	minX = 1e6; maxX = -1e6;
	minY = 1e6; maxY = -1e6;
	minZ = 1e6; maxZ = -1e6;

        radiusX=0;
        radiusY=0;
        radiusZ=0;

	fCenterX=0;
	fCenterY=0;
	fCenterZ=0;

        lTotalTriangles=0;
        lTotalVertexes=0;

        while(ftrisearch(in,"Tri-mesh")==0)
        {
	    LONG i;
                                                 //search for object data
            fsearch(in,"Vertices:");             //get VertexNUM
            fscanf(in,"%d",&nv);
            fsearch(in,"Faces:");                //get PolyNUM
            fscanf(in,"%d",&np);

            fsearch(in,"Vertex"); fsearch(in,"list:");
            memset(SubObjectTable, 0, sizeof(SubObjectTable));
            for (i=0; i<nv; i++) {
                fsearch(in,"Vertex");
                fscanf(in,"%d",&vertexNR);
                fsearchar(in,':');
                fsearchar(in,':');
                fscanf(in,"%f",&fVertexX);
                fsearchar(in,':');
                fscanf(in,"%f",&fVertexY);
                fsearchar(in,':');
                fscanf(in,"%f",&fVertexZ);

                if (i>=MAX_VERTEXES_OF_SUBOBJECTS)
                    throw new RendererException(
		    	"Out of SubObject Vertexes in Object construction");

		if ((fScale*fVertexX)>maxX)
			maxX = (fScale*fVertexX);
		if ((fScale*fVertexX)<minX)
			minX = (fScale*fVertexX);

		if ((fScale*fVertexY)>maxY)
			maxY = (fScale*fVertexY);
		if ((fScale*fVertexY)<minY)
			minY = (fScale*fVertexY);

		if ((fScale*fVertexZ)>maxZ)
			maxZ = (fScale*fVertexZ);
		if ((fScale*fVertexZ)<minZ)
			minZ = (fScale*fVertexZ);

                SubObjectTable[i] = new Vertex(this, fScale*fVertexX,
                                                     fScale*fVertexY,
                                                     fScale*fVertexZ);

            }

            if (!final_pass) {
                for(i=0; i<nv; i++)
                    delete SubObjectTable[i];
            }
            fsearch(in,"Face"); fsearch(in,"list:");
            for (i=0; i<np; i++) {
                Material *node;

                fsearch(in,"Face");
                fscanf(in,"%d",&polyNR);
                if (polyNR != i) {
#ifdef LOG
                    fclose(fpout);
#endif
                    cout << "Error:\ni=" << i << ", polyNR=" << polyNR << "\n";
                    cout << "pointA=" << pointA << ",pointB=" << pointB << ",pointC=" << pointC << "\n";
                    cout << "material" << material << ",tmpSTR=" << tmpSTR << "\n";
                    throw new RendererException("Error in Reading Faces");
                }
                fsearchar(in,':');
                fsearchar(in,':');
                fscanf(in,"%d",&pointA);
                fsearchar(in,':');
                fscanf(in,"%d",&pointB);
                fsearchar(in,':');
                fscanf(in,"%d",&pointC);

//                fscanf(in,"%[^\n]%*9s",tmpSTR);
                fsearchar(in,':');
                fsearchar(in,'"');
//                getqstring(in, material);
                fscanf(in,"%[^\"]", tmpSTR);
                sprintf(material, "\"%s\"", tmpSTR);
#ifdef LOG
                fprintf(fpout, "Face: %d = %d %d %d %s\n",
                    polyNR, pointA, pointB, pointC, material);
#endif
                if (final_pass) {
                    BOOLEAN found=FALSE;

                    if (
                        (material[0]=='"') &&
                        (tolower(material[1])=='r') &&
                        (isdigit(material[2]))
                       )
                    {
                        node=GetHeadMaterial();

                        while(node!=NULL) {

                            if(strcasecmp(material, node->GetMaterialName()) == 0) {
                                found = TRUE;
                                break;
                            }
                            else
                                node = node->GetNextNode();
                        }
                        if(!found) {
                            node = new Material(this, material);

                        }
                    }
                    else
                        node = white;

                }

                if ((final_pass) &&
                   ((!SubObjectTable[pointA]) ||
                    (!SubObjectTable[pointB]) ||
                    (!SubObjectTable[pointC])))
                    throw new RendererException("Oh no!\n");

                if (final_pass) {
                    Triangle *tmp = new Triangle(this, SubObjectTable[pointA],
                                                       SubObjectTable[pointB],
                                                       SubObjectTable[pointC],
                                                       node);
		    if(!tmp)
			throw new RendererException("Oh! no memory!\n");
                }
            }
            lTotalVertexes+=nv;
            lTotalTriangles+=np;

            if (final_pass) {
                cout << "Vertexes: " << lTotalVertexes;
                cout << " Triangles: " << lTotalTriangles << "\n";
            }
        }
        fclose(in);

	radiusX = maxX-minX;
	radiusY = maxY-minY;
	radiusZ = maxZ-minZ;

	fCenterX = minX + radiusX/2;
	fCenterY = minY + radiusY/2;
	fCenterZ = minZ + radiusZ/2;

        if(final_pass)
            break;

        FLOAT max_radius = radiusX;
        if (radiusY>max_radius) max_radius = radiusY;
        if (radiusZ>max_radius) max_radius = radiusZ;

        if ((max_radius>250) || (max_radius<225)) {
/*
            cout << "Radius of objects: " << max_radius << "\n";
            cout << "Please use scale factor: " << fScale*250.0/max_radius << "\n";
            throw new RendererException("Objects must have \n\t400<radius<500\n");
*/
            fScale = fScale*250.0/max_radius;
            cout << "Recalculating object with fScale: " << fScale << "\n";
        }
        else {
            fScale = fScale*250.0/max_radius;
            final_pass = TRUE;
            cout << "Final pass with fScale: " << fScale << "\n";
            SetHeadVertex(NULL);
            SetHeadTriangle(NULL);
        }
    }
#ifdef LOG
    fclose(fpout);
#endif
}


void Object3D::fix_normals(void)
{
    FLOAT aax, aay, aaz, bbx, bby, bbz;
    FLOAT dumx, dumy, dumz;
    FLOAT fSumm;

    Triangle *tri;

    LONG lTrianglesWithNoNormal = 0;
    for(tri=headTriangle; tri!=NULL;) {
        Vertex *A = tri->GetV0();
        Vertex *B = tri->GetV1();
        Vertex *C = tri->GetV2();

        aax = B->GetX() - A->GetX();
        aay = B->GetY() - A->GetY();
        aaz = B->GetZ() - A->GetZ();

        bbx = C->GetX() - A->GetX();
        bby = C->GetY() - A->GetY();
        bbz = C->GetZ() - A->GetZ();

        dumx=aay*bbz-bby*aaz;
        dumy=bbx*aaz-aax*bbz;
        dumz=aax*bby-aay*bbx;

	fSumm=sqrt(dumx*dumx+dumy*dumy+dumz*dumz);
        if (fSumm>1e-5) {
	    
	    dumx = dumx/fSumm;
	    dumy = dumy/fSumm;
	    dumz = dumz/fSumm;

	}
	else {
	    lTrianglesWithNoNormal++;
#if 0
	    printf("ooooops!!! | %f %f %f | = %f\n",
	    		dumx, dumy, dumz, fSumm);
	    printf("A: %d %d %d\n", A->GetX(), A->GetY(), A->GetZ());
	    printf("B: %d %d %d\n", B->GetX(), B->GetY(), B->GetZ());
	    printf("C: %d %d %d\n", C->GetX(), C->GetY(), C->GetZ());

	    exit(-1);
#endif
	}

        tri->SetNormals(dumx, dumy, dumz);

        A->AddToNormal(dumx, dumy, dumz);
        B->AddToNormal(dumx, dumy, dumz);
        C->AddToNormal(dumx, dumy, dumz);

        tri = tri->GetNextNode();
    }

    if (lTrianglesWithNoNormal) {
	printf(	"%d triangles could not produce a normal...\n",
		lTrianglesWithNoNormal);
    }
    Vertex *v;
    LONG lCountz=0;

    for(v=headVertex; v!=NULL;) {
        fSumm=sqrt(( (v->GetNormalX()) * (v->GetNormalX()) +
                     (v->GetNormalY()) * (v->GetNormalY()) +
                     (v->GetNormalZ()) * (v->GetNormalZ()) ));
        if (fSumm<1e-5) {
            lCountz++;
            fSumm=1.0;
            v->SetAlways();
        }
        v->SetNormals( 	v->GetNormalX() / fSumm,
			v->GetNormalY() / fSumm,
			v->GetNormalZ() / fSumm);


        v=v->GetNextNode();
    }
    if(lCountz) {
        cout << "I found " << lCountz << " non-colorable polygons.\n";
    }
}


void Object3D::RenderDots(Camera *camera)
{
    Vertex *node;
 
    for(node=headVertex; node!=NULL;) {
        node->Transform(camera);
	if (node->GetTransformedZ()>ZCLIP) {
	    plot(
		(LONG)(MAXX/2 + 
			FOV*node->GetTransformedY()/node->GetTransformedZ()),
		(LONG)(MAXY/2 - 
			FOV*node->GetTransformedX()/node->GetTransformedZ()),
	    255);
	}
        node = node->GetNextNode();
    }
}


void Object3D::RenderWireFrame(Camera *camera)
{
    Vertex *node;
    Triangle *tri;
    
    for(node=headVertex; node!=NULL;) {
        node->Transform(camera);
//        plot( node->GetTransformedX(), node->GetTransformedY(), 255);
        node = node->GetNextNode();
    }

    for(tri=headTriangle; tri!=NULL; tri = tri->GetNextNode()) {

        FLOAT eex, eey, eez;

	if((tri->GetV0()->GetTransformedZ()<ZCLIP) ||
	   (tri->GetV1()->GetTransformedZ()<ZCLIP) ||
	   (tri->GetV2()->GetTransformedZ()<ZCLIP))
	   continue;

	eex = camera->GetEyeX() - tri->GetCenterX();
	eey = camera->GetEyeY() - tri->GetCenterY();
	eez = camera->GetEyeZ() - tri->GetCenterZ();
	if
        (

            (
                eex * tri->GetNormalX() +
	        eey * tri->GetNormalY() +
	        eez * tri->GetNormalZ() >=0
            )

/*
            ||
            (
                tri->GetV0()->IsAlways() ||
                tri->GetV1()->IsAlways() ||
                tri->GetV2()->IsAlways()
            )
*/
        )
        {
	    FLOAT x0,y0,z0, x1,y1,z1, x2,y2,z2;

	    x0 = tri->GetV0()->GetTransformedX();
	    y0 = tri->GetV0()->GetTransformedY();
	    z0 = tri->GetV0()->GetTransformedZ();
	    if(z0<0)
	    	continue;
	    x1 = tri->GetV1()->GetTransformedX();
	    y1 = tri->GetV1()->GetTransformedY();
	    z1 = tri->GetV1()->GetTransformedZ();
	    if(z1<0)
	    	continue;
	    x2 = tri->GetV2()->GetTransformedX();
	    y2 = tri->GetV2()->GetTransformedY();
	    z2 = tri->GetV2()->GetTransformedZ();
	    if(z2<0)
	    	continue;

            WuLine(
		(MAXX/2.0) + FOV*y0/z0,
		(MAXY/2.0) - FOV*x0/z0,
		(MAXX/2.0) + FOV*y1/z1,
		(MAXY/2.0) - FOV*x1/z1,
                tri->GetMaterial()->GetPaletteColor()
            );
            WuLine(
		(MAXX/2.0) + FOV*y1/z1,
		(MAXY/2.0) - FOV*x1/z1,
		(MAXX/2.0) + FOV*y2/z2,
		(MAXY/2.0) - FOV*x2/z2,
                tri->GetMaterial()->GetPaletteColor()
            );
            WuLine(
		(MAXX/2.0) + FOV*y2/z2,
		(MAXY/2.0) - FOV*x2/z2,
		(MAXX/2.0) + FOV*y0/z0,
		(MAXY/2.0) - FOV*x0/z0,
                tri->GetMaterial()->GetPaletteColor()
            );
#ifdef DEBUG
            char c=(char)getch();
            if (c==27) {
                close256();
                cout << "I stopped at the following polygon: \n";
                cout << "Vertex0 : " << tri->GetV0()->GetX();
                cout << tri->GetV0()->GetY() << tri->GetV0()->GetZ() << "\n";
                cout << "Vertex1 : " << tri->GetV1()->GetX();
                cout << tri->GetV1()->GetY() << tri->GetV1()->GetZ() << "\n";
                cout << "Vertex2 : " << tri->GetV2()->GetX();
                cout << tri->GetV2()->GetY() << tri->GetV2()->GetZ() << "\n";
                exit(0);
            }
#endif

        }
    }
}


void Object3D::RenderDepthWireFrame(Camera *camera)
{
    Vertex *node;
    Triangle *tri;
    
    for(node=headVertex; node!=NULL;) {
        node->Transform(camera);
        node = node->GetNextNode();
    }

    LONG lDist, lDistMin=2100000000, lDistMax=0;
    LONG lVisibleTriangles;

    if(!sortArray)
        sortArray = (Triangle **)malloc(lTotalTriangles*sizeof(Triangle*));

    memset(countArray, 0, MAXZDEPTH*sizeof(LONG));
    lVisibleTriangles = 0;

    for(tri=headTriangle; tri!=NULL; tri = tri->GetNextNode()) {

        FLOAT eex, eey, eez;

	if ((tri->GetV0()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV1()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV2()->GetTransformedZ()<ZCLIP)) {
	    tri->SetVisible(FALSE);
	    continue;
       	}

        eex = camera->GetEyeX() - tri->GetCenterX();
	eey = camera->GetEyeY() - tri->GetCenterY();
	eez = camera->GetEyeZ() - tri->GetCenterZ();

	if
        (
            (
                tri->GetV0()->IsAlways() ||
                tri->GetV1()->IsAlways() ||
                tri->GetV2()->IsAlways()
            )
            ||
            (
                eex * tri->GetNormalX() +
	        eey * tri->GetNormalY() +
	        eez * tri->GetNormalZ() >=0
            )
        )
        {
            tri->SetZDistance(lDist=(LONG)((tri->GetV0()->GetZDistance() +
                                            tri->GetV1()->GetZDistance() +
                                            tri->GetV2()->GetZDistance()) / 3));
            if (lDist>MAXZDEPTH-1) {
                close256();
                throw new RendererException("Out of radius in sorting table.");
            }

            if (lDistMax < lDist)
                lDistMax = lDist;

            if (lDistMin > lDist)
                lDistMin = lDist;

            countArray[lDist]++;
            tri->SetVisible(TRUE);
            lVisibleTriangles++;
        }
        else
            tri->SetVisible(FALSE);
    }

    for(LONG i=1; i<MAXZDEPTH; i++)
        countArray[i]+=countArray[i-1];

    for(tri=headTriangle; tri!=NULL;) {
        if (tri->IsVisible()) {
            unsigned offset = countArray[ lDist = tri->GetZDistance() ]-1;

            countArray[ lDist ]--;
            sortArray[offset] = tri;
        }
        tri = tri->GetNextNode();
    }

    for(LONG iTri=lVisibleTriangles-1; iTri>=0; iTri--) {

        tri = sortArray[iTri];
    
	FLOAT x0,y0,z0, x1,y1,z1, x2,y2,z2;

	x0 = tri->GetV0()->GetTransformedX();
	y0 = tri->GetV0()->GetTransformedY();
	z0 = tri->GetV0()->GetTransformedZ();
	if(z0<0)
	    continue;
	x1 = tri->GetV1()->GetTransformedX();
	y1 = tri->GetV1()->GetTransformedY();
	z1 = tri->GetV1()->GetTransformedZ();
	if(z1<0)
	    continue;
	x2 = tri->GetV2()->GetTransformedX();
	y2 = tri->GetV2()->GetTransformedY();
	z2 = tri->GetV2()->GetTransformedZ();
	if(z2<0)
	    continue;

#define DDEPTHH 255
	SolidLine(
	    (MAXX/2.0) + FOV*y0/z0,
	    (MAXY/2.0) - FOV*x0/z0,
	    (MAXX/2.0) + FOV*y1/z1,
	    (MAXY/2.0) - FOV*x1/z1,
            (unsigned char)
            (DDEPTHH-(unsigned char)((tri->GetZDistance()-lDistMin)*DDEPTHH/(lDistMax-lDistMin)))
	);
	SolidLine(
	    (MAXX/2.0) + FOV*y1/z1,
	    (MAXY/2.0) - FOV*x1/z1,
	    (MAXX/2.0) + FOV*y2/z2,
	    (MAXY/2.0) - FOV*x2/z2,
            (unsigned char)
            (DDEPTHH-(unsigned char)((tri->GetZDistance()-lDistMin)*DDEPTHH/(lDistMax-lDistMin)))
	);
	SolidLine(
	    (MAXX/2.0) + FOV*y2/z2,
	    (MAXY/2.0) - FOV*x2/z2,
	    (MAXX/2.0) + FOV*y0/z0,
	    (MAXY/2.0) - FOV*x0/z0,
            (unsigned char)
            (DDEPTHH-(unsigned char)((tri->GetZDistance()-lDistMin)*DDEPTHH/(lDistMax-lDistMin)))
	);

#ifdef DEBUG
        char c=(char)getch();
        if (c==27) {
            close256();
            cout << "I stopped at the following polygon: \n";
            cout << "Vertex0 : " << tri->GetV0()->GetX();
            cout << tri->GetV0()->GetY() << tri->GetV0()->GetZ() << "\n";
            cout << "Vertex1 : " << tri->GetV1()->GetX();
            cout << tri->GetV1()->GetY() << tri->GetV1()->GetZ() << "\n";
            cout << "Vertex2 : " << tri->GetV2()->GetX();
            cout << tri->GetV2()->GetY() << tri->GetV2()->GetZ() << "\n";
            exit(0);
        }
#endif
    }
}



void Object3D::RenderFlatWire(Camera *camera, BOOLEAN bWireOn)
{
    Vertex *node;
    Triangle *tri;
    
    for(node=headVertex; node!=NULL;) {
        node->Transform(camera);
//        plot( node->GetTransformedX(), node->GetTransformedY(), 255);
        node = node->GetNextNode();
    }

    LPOLY p;
    LPOINT po1, po2, po3;
    LONG lDist;
    LONG lVisibleTriangles;

    if(!sortArray)
        sortArray = (Triangle **)malloc(lTotalTriangles*sizeof(Triangle*));

    memset(countArray, 0, MAXZDEPTH*sizeof(LONG));
    lVisibleTriangles = 0;

    p.P1=&po1; p.P2=&po2; p.P3=&po3;

    for(tri=headTriangle; tri!=NULL; tri=tri->GetNextNode()) {

	if ((tri->GetV0()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV1()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV2()->GetTransformedZ()<ZCLIP)) {
	    tri->SetVisible(FALSE);
	    continue;
       	}

        FLOAT eex, eey, eez;

	eex = camera->GetEyeX() - tri->GetCenterX();
	eey = camera->GetEyeY() - tri->GetCenterY();
	eez = camera->GetEyeZ() - tri->GetCenterZ();

	if
        (
            (
                tri->GetV0()->IsAlways() ||
                tri->GetV1()->IsAlways() ||
                tri->GetV2()->IsAlways()
            )
            ||
            (
                eex * tri->GetNormalX() +
	        eey * tri->GetNormalY() +
	        eez * tri->GetNormalZ() >=0
            )
        )
        {
            tri->SetZDistance(lDist=(LONG)((tri->GetV0()->GetZDistance() +
				            tri->GetV1()->GetZDistance() +
					    tri->GetV2()->GetZDistance())/3));
            if (lDist>MAXZDEPTH-1) {
                close256();
                throw new RendererException("Out of radius in sorting table.");
            }
            countArray[lDist]++;
            tri->SetVisible(TRUE);
            lVisibleTriangles++;
        }
        else
            tri->SetVisible(FALSE);
    }

    for(LONG i=1; i<MAXZDEPTH; i++)
        countArray[i]+=countArray[i-1];

    for(tri=headTriangle; tri!=NULL;) {
        if (tri->IsVisible()) {
            unsigned offset = countArray[ lDist = tri->GetZDistance() ]-1;

            countArray[ lDist ]--;
            sortArray[offset] = tri;
        }
        tri = tri->GetNextNode();
    }

    for(LONG iTri=lVisibleTriangles-1; iTri>=0; iTri--) {

        tri = sortArray[iTri];

	FLOAT x0,y0,z0, x1,y1,z1, x2,y2,z2;

	x0 = tri->GetV0()->GetTransformedX();
	y0 = tri->GetV0()->GetTransformedY();
	z0 = tri->GetV0()->GetTransformedZ();
	if(z0<0)
	    continue;
	x1 = tri->GetV1()->GetTransformedX();
	y1 = tri->GetV1()->GetTransformedY();
	z1 = tri->GetV1()->GetTransformedZ();
	if(z1<0)
	    continue;
	x2 = tri->GetV2()->GetTransformedX();
	y2 = tri->GetV2()->GetTransformedY();
	z2 = tri->GetV2()->GetTransformedZ();
	if(z2<0)
	    continue;

        po1.X = LONG((MAXX/2.0) + FOV*y0/z0);
        po1.Y = LONG((MAXY/2.0) - FOV*x0/z0);

        po2.X = LONG((MAXX/2.0) + FOV*y1/z1);
        po2.Y = LONG((MAXY/2.0) - FOV*x1/z1);

        po3.X = LONG((MAXX/2.0) + FOV*y2/z2);
        po3.Y = LONG((MAXY/2.0) - FOV*x2/z2);

        po1.Color = tri->GetMaterial()->GetPaletteColor();
        po2.Color = tri->GetMaterial()->GetPaletteColor();
        po3.Color = tri->GetMaterial()->GetPaletteColor();

        GouraudFill_Tri(&p);

	if (bWireOn) {
	    SolidLine(
		(MAXX/2.0) + FOV*y0/z0,
		(MAXY/2.0) - FOV*x0/z0,
		(MAXX/2.0) + FOV*y1/z1,
		(MAXY/2.0) - FOV*x1/z1,
		255
	    );
	    SolidLine(
		(MAXX/2.0) + FOV*y1/z1,
		(MAXY/2.0) - FOV*x1/z1,
		(MAXX/2.0) + FOV*y2/z2,
		(MAXY/2.0) - FOV*x2/z2,
		255
	    );
	    SolidLine(
		(MAXX/2.0) + FOV*y2/z2,
		(MAXY/2.0) - FOV*x2/z2,
		(MAXX/2.0) + FOV*y0/z0,
		(MAXY/2.0) - FOV*x0/z0,
		255
	    );
	}


#ifdef DEBUG
        char c=(char)getch();
        if (c==27) {
            close256();
            cout << "I stopped at the following polygon: \n";
            cout << "Vertex0 : " << tri->GetV0()->GetX();
            cout << tri->GetV0()->GetY() << tri->GetV0()->GetZ() << "\n";
            cout << "Vertex1 : " << tri->GetV1()->GetX();
            cout << tri->GetV1()->GetY() << tri->GetV1()->GetZ() << "\n";
            cout << "Vertex2 : " << tri->GetV2()->GetX();
            cout << tri->GetV2()->GetY() << tri->GetV2()->GetZ() << "\n";
            exit(0);
        }
#endif

    }
}

void Object3D::RenderFastGouraud(Camera *camera, Light *light)
{
    Vertex *node;
    Triangle *tri;
    
    if (light)
        for(node=headVertex; node!=NULL;) {
            node->Transform(camera);
            node->Illuminate(light, camera);
            node = node->GetNextNode();
        }
    else
        for(node=headVertex; node!=NULL;) {
            node->Transform(camera);
            node = node->GetNextNode();
        }

    LPOLY p;
    LPOINT po1, po2, po3;
    LONG lDist;
    LONG lVisibleTriangles;

    if(!sortArray)
        sortArray = (Triangle **)malloc(lTotalTriangles*sizeof(Triangle*));

    memset(countArray, 0, MAXZDEPTH*sizeof(LONG));
    lVisibleTriangles = 0;

    p.P1=&po1; p.P2=&po2; p.P3=&po3;

    for(tri=headTriangle; tri!=NULL; tri=tri->GetNextNode()) {
    
	if ((tri->GetV0()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV1()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV2()->GetTransformedZ()<ZCLIP)) {
	    tri->SetVisible(FALSE);
	    continue;
       	}

        FLOAT eex, eey, eez;

	eex = camera->GetEyeX() - tri->GetCenterX();
	eey = camera->GetEyeY() - tri->GetCenterY();
	eez = camera->GetEyeZ() - tri->GetCenterZ();

	if
        (
            (
                tri->GetV0()->IsAlways() ||
                tri->GetV1()->IsAlways() ||
                tri->GetV2()->IsAlways()
            )
            ||
            (
                eex * tri->GetNormalX() +
	        eey * tri->GetNormalY() +
	        eez * tri->GetNormalZ() >=0
            )
        )
        {
            tri->SetZDistance(lDist=(LONG)((tri->GetV0()->GetZDistance() +
				            tri->GetV1()->GetZDistance() +
					    tri->GetV2()->GetZDistance())/3));
            if (lDist>MAXZDEPTH-1) {
                close256();
                throw new RendererException("Out of radius in sorting table.");
            }
            countArray[lDist]++;
            tri->SetVisible(TRUE);
            lVisibleTriangles++;
        }
        else
            tri->SetVisible(FALSE);
    }

    for(LONG i=1; i<MAXZDEPTH; i++)
        countArray[i]+=countArray[i-1];

    for(tri=headTriangle; tri!=NULL;) {
        if (tri->IsVisible()) {
            unsigned offset = countArray[ lDist = tri->GetZDistance() ]-1;

            countArray[ lDist ]--;
            sortArray[offset] = tri;
        }
        tri = tri->GetNextNode();
    }

    for(LONG iTri=lVisibleTriangles-1; iTri>=0; iTri--) {

        tri = sortArray[iTri];

	FLOAT x0,y0,z0, x1,y1,z1, x2,y2,z2;

	x0 = tri->GetV0()->GetTransformedX();
	y0 = tri->GetV0()->GetTransformedY();
	z0 = tri->GetV0()->GetTransformedZ();
	if(z0<0)
	    continue;
	x1 = tri->GetV1()->GetTransformedX();
	y1 = tri->GetV1()->GetTransformedY();
	z1 = tri->GetV1()->GetTransformedZ();
	if(z1<0)
	    continue;
	x2 = tri->GetV2()->GetTransformedX();
	y2 = tri->GetV2()->GetTransformedY();
	z2 = tri->GetV2()->GetTransformedZ();
	if(z2<0)
	    continue;

        po1.X = LONG((MAXX/2.0) + FOV*y0/z0);
        po1.Y = LONG((MAXY/2.0) - FOV*x0/z0);

        po2.X = LONG((MAXX/2.0) + FOV*y1/z1);
        po2.Y = LONG((MAXY/2.0) - FOV*x1/z1);

        po3.X = LONG((MAXX/2.0) + FOV*y2/z2);
        po3.Y = LONG((MAXY/2.0) - FOV*x2/z2);

	po1.Color = tri->GetV0()->GetIntensity();
	po2.Color = tri->GetV1()->GetIntensity();
	po3.Color = tri->GetV2()->GetIntensity();

        GouraudFill_Tri(&p);
    }
}

extern FLOAT ZBuffer[MAXY][MAXX];

void Object3D::RenderZGouraud(Camera *camera, Light *light)
{
    Vertex *node;
    Triangle *tri;
    
    if (light)
        for(node=headVertex; node!=NULL;) {
            node->Transform(camera);
            node->Illuminate(light, camera);
            node = node->GetNextNode();
        }
    else
        for(node=headVertex; node!=NULL;) {
            node->Transform(camera);
            node->SetIntensity(128);
            node = node->GetNextNode();
        }

    FPOLY p;
    FPOINT po1, po2, po3;

    p.P1=&po1; p.P2=&po2; p.P3=&po3;

    for(int amanx=0; amanx<MAXX; amanx++)
	for(int amany=0; amany<MAXY; amany++)
	    ZBuffer[amany][amanx]=1E-30;

    for(tri=headTriangle; tri!=NULL; tri=tri->GetNextNode()) {

	if ((tri->GetV0()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV1()->GetTransformedZ()<ZCLIP) ||
	    (tri->GetV2()->GetTransformedZ()<ZCLIP)) {
	    continue;
       	}

        FLOAT eex, eey, eez;

	eex = camera->GetEyeX() - tri->GetCenterX();
	eey = camera->GetEyeY() - tri->GetCenterY();
	eez = camera->GetEyeZ() - tri->GetCenterZ();

	if ((tri->GetV0()->IsAlways() ||
	     tri->GetV1()->IsAlways() ||
	     tri->GetV2()->IsAlways())
            ||
            (eex * tri->GetNormalX() +
	     eey * tri->GetNormalY() +
	     eez * tri->GetNormalZ() >=0))
        {
	    FLOAT x0,y0,z0, x1,y1,z1, x2,y2,z2;

	    x0 = tri->GetV0()->GetTransformedX();
	    y0 = tri->GetV0()->GetTransformedY();
	    z0 = tri->GetV0()->GetTransformedZ();
	    x1 = tri->GetV1()->GetTransformedX();
	    y1 = tri->GetV1()->GetTransformedY();
	    z1 = tri->GetV1()->GetTransformedZ();
	    x2 = tri->GetV2()->GetTransformedX();
	    y2 = tri->GetV2()->GetTransformedY();
	    z2 = tri->GetV2()->GetTransformedZ();

	    po1.X = (MAXX/2) + FOV*y0/z0;
	    po1.Y = (MAXY/2) - FOV*x0/z0;
	    po1.Z = 1.0/z0;

	    po2.X = (MAXX/2) + FOV*y1/z1;
	    po2.Y = (MAXY/2) - FOV*x1/z1;
	    po2.Z = 1.0/z1;

	    po3.X = (MAXX/2) + FOV*y2/z2;
	    po3.Y = (MAXY/2) - FOV*x2/z2;
	    po3.Z = 1.0/z2;

	    po1.Color = tri->GetV0()->GetIntensity();
	    po2.Color = tri->GetV1()->GetIntensity();
	    po3.Color = tri->GetV2()->GetIntensity();

	    ZBufferTri(&p);
        }
    }
}


Object3D::~Object3D()
{
//    cout << "I should free everything here.";
}
