/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "Vgalib.h"
#include "Defines.h"
#include "Wu.h"

#include <SDL/SDL.h>

SDL_Surface *screen = NULL;

unsigned offst[MAXY];
UCHAR *_w_screen = NULL;

void clear_page(void)
{
    memset(_w_screen, 0, MAXX*MAXY);
}

void show_page()
{
    if ( SDL_MUSTLOCK(screen) ) {
	SDL_UnlockSurface(screen);
    }
    SDL_UpdateRect(screen, 0,0,MAXX,MAXY);
    if ( SDL_MUSTLOCK(screen) ) {
	if ( SDL_LockSurface(screen) < 0 ) {
	    fprintf(stderr, "Couldn't lock screen: %s\n",
		SDL_GetError());
	    exit(1);
	}
    }
    _w_screen = (Uint8 *) screen->pixels;
}

int oldmode;

void init256()
{
    register int k1;
    
    for (k1=0; k1<=MAXY-1; k1++)
        offst[k1] = k1*MAXX;

    /* Initialize the SDL library */
    if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
	fprintf(stderr,
	    "Couldn't initialize SDL: %s\n", SDL_GetError());
	exit(1);
    }

    /* Clean up on exit */
    atexit(SDL_Quit);

    /* Initialize the display - palettized mode */
#ifdef FULLSCREEN
    screen = SDL_SetVideoMode(MAXX, MAXY, 8, SDL_HWSURFACE | SDL_FULLSCREEN);
#else
    screen = SDL_SetVideoMode(MAXX, MAXY, 8, SDL_HWSURFACE | SDL_HWPALETTE);
#endif
    if ( screen == NULL ) {
	fprintf(stderr, "Couldn't set video mode: %s\n",
	    SDL_GetError());
	exit(1);
    }

    if ( SDL_MUSTLOCK(screen) ) {
	if ( SDL_LockSurface(screen) < 0 ) {
	    fprintf(stderr, "Couldn't lock screen: %s\n",
		SDL_GetError());
	    exit(1);
	}
    }

    _w_screen = (Uint8 *) screen->pixels;
    if(!_w_screen)
    {
        fprintf(stderr, "Screen buffer can't be allocated!");
        return;
    }
}

void close256()
{
    if ( SDL_MUSTLOCK(screen) ) {
	SDL_UnlockSurface(screen);
    }
}

void plot(int r,int k,int l)
{
    register unsigned char *s=(unsigned char *)_w_screen;
    
    if ((r<MAXX) && (r>0) && (k>0) && (k<MAXY))
        *(s+offst[k]+r)=(unsigned char)l;
}

void qplot(int r,int k,int l)
{
    register unsigned char *s=(unsigned char *)_w_screen;
    
    *(s+offst[k]+r)=(unsigned char)l;
}

void waitkey()
{
//    int c;

//    if((c=getch())==0) c=getch();
}

SDL_Color palette[256];

void set256Palette(unsigned char *s)
{
    int i;

    for(i=0; i<256; i++) {
    	palette[i].r = *s++;
    	palette[i].g = *s++;
    	palette[i].b = *s++;
    }
    SDL_SetColors(screen, palette, 0, 256);
}

void pal()
{
    int i;

    for(i=0; i<256; i++) {
    	palette[i].r = i;
    	palette[i].g = i;
    	palette[i].b = i;
    }
    SDL_SetColors(screen, palette, 0, 256);
}

void vgaline(int x1,int y1,int x2,int y2,unsigned char colour, unsigned char *page = _w_screen)
{
    int xtmp, ytmp, i, j, slope, start, xold, xnew;
    
    if (y1>y2) {
        ytmp=y1; xtmp=x1;
        y1=y2; x1=x2;
        y2=ytmp; x2=xtmp;
    }
    if (y2==y1) {
        if (x2>=x1)
            memset(page+offst[y1]+x1,colour,x2-x1);
        else
            memset(page+offst[y1]+x2,colour,x1-x2);
        return;
    }
    slope=((x2-x1) << 16) /(y2-y1+1);
    start=x1 << 16;
    *(page+offst[y1] + x1)=colour;
    xold = x1;
    if(x1<x2) {
        for (i=y1; i<=y2; i++) {
            start+=slope;
            xnew = start >> 16; //*((short *) ((char *)&start+2));
            if (xold==xnew)
                *(page+offst[i] + xnew)=colour;
            else
              for(j=xold; j<xnew; j++)
                *(page+offst[i] + j)=colour;
            xold = xnew;
        }
    }
    else if (x1>x2) {
        for (i=y1; i<=y2; i++) {
            start+=slope;
            xnew = start >> 16; //*((short *) ((char *)&start+2));
            if (xold==xnew)
                *(page+offst[i] + xnew)=colour;
            else
              for(j=xold; j>xnew; j--)
                *(page+offst[i] + j)=colour;
            xold = xnew;
        }
    }
    else {
        for (i=y1+1; i<y2; i++)
            *(page+offst[i] + x1)=colour;
    }
}

void code(FLOAT x,FLOAT y,unsigned char *c)
{
    *c=0;
    if (x<0) *c=*c | (unsigned char)1;
    else if (x>MAXX-1) *c=*c | (unsigned char)2;
    if (y<0) *c=*c | (unsigned char)4;
    else if (y>MAXY-1) *c=*c | (unsigned char)8;
}

void WuLine(FLOAT x1,FLOAT y1,FLOAT x2,FLOAT y2,unsigned char colour)
{
    unsigned char c,c1,c2;
    FLOAT x,y;
    
    code(x1,y1,&c1);code(x2,y2,&c2);
    while ((c1!=0) || (c2!=0)) {
        if ((c1 & c2)!=0) return;
        c=c1;
        if (c==0) c=c2;
        if ((c & 1)!=0)
        {
            y=(y1+((y2-y1)*(-x1)/(x2-x1)));
            x=0;
        }
        else if ((c & 2)!=0)
        {
            y=(y1+((y2-y1)*(MAXX-1-x1)/(x2-x1)));
            x=MAXX-1;
        }
        else if ((c & 8)!=0)
        {
            x=(x1+((x2-x1)*(MAXY-1-y1)/(y2-y1)));
            y=MAXY-1;
        }
        else if ((c & 4)!=0)
        {
            x=(x1+((x2-x1)*(-y1)/(y2-y1)));
            y=0;
        }
        if (c==c1)
        {
            x1=x;y1=y;
            code(x1,y1,&c1);
        }
        else
        {
            x2=x;y2=y;
            code(x2,y2,&c2);
        }
    }
    DrawWuLine(
    	_w_screen,
	(int) x1,(int) y1,(int) x2,(int) y2,
	0,
	64,
	6);
}

void SolidLine(FLOAT x1,FLOAT y1,FLOAT x2,FLOAT y2,unsigned char colour)
{
    unsigned char c,c1,c2;
    FLOAT x,y;
    
    code(x1,y1,&c1);code(x2,y2,&c2);
    while ((c1!=0) || (c2!=0)) {
        if ((c1 & c2)!=0) return;
        c=c1;
        if (c==0) c=c2;
        if ((c & 1)!=0)
        {
            y=(y1+((y2-y1)*(-x1)/(x2-x1)));
            x=0;
        }
        else if ((c & 2)!=0)
        {
            y=(y1+((y2-y1)*(MAXX-1-x1)/(x2-x1)));
            x=MAXX-1;
        }
        else if ((c & 8)!=0)
        {
            x=(x1+((x2-x1)*(MAXY-1-y1)/(y2-y1)));
            y=MAXY-1;
        }
        else if ((c & 4)!=0)
        {
            x=(x1+((x2-x1)*(-y1)/(y2-y1)));
            y=0;
        }
        if (c==c1)
        {
            x1=x;y1=y;
            code(x1,y1,&c1);
        }
        else
        {
            x2=x;y2=y;
            code(x2,y2,&c2);
        }
    }
    vgaline( (int) x1,(int) y1,(int) x2,(int) y2, colour);
}
