/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <assert.h>

#include "Vgalib.h"
#include "Gouraud.h"
#include "Defines.h"

//#define LOG
/***************************
#define INIT(var, initial, final, steps) \
    int var##dword = ((int)initial) << 16;  \
    int var##step  = steps?((((int)final)-((int)initial))<<16)/((int)steps):0;
#ifdef LOG
    fprintf(log, "%s start: %d end:%d in %d steps\n", \
       #var, initial, final, steps); \
    fprintf(log, "i.e. %sdword:%d, %sstep:%d\n",  \
       #var, var##dword, #var, var##step);
#endif

#define VALUE(var)  *(((short *) & var##dword)+1)

#define NEXT(var)  var##dword += var##step
******************/

#define INIT(var, initial, final, steps) \
    FLOAT var##float = (FLOAT)initial;  \
    FLOAT var##step  = ((LONG)steps)?((((FLOAT)final)-((FLOAT)initial)))/((LONG)steps):0; \

#define VALUE(var)  var##float

#define NEXT(var)  var##float += var##step


#define WIDTH  MAXX
#define HEIGHT MAXY

struct PointData {
    int On;
    FLOAT x, c;
    FLOAT z;
};

struct ScanLine {
    int Active;
    PointData Left, Right;
};

FLOAT ZBuffer[HEIGHT][WIDTH];

ScanLine scanlines[HEIGHT];

#ifdef LOG
FILE *log;
#endif

void inline plotZPoint(FLOAT xx, FLOAT yy, FLOAT z, FLOAT cc)
{
    int x = (int)xx;
    int y = (int)yy;
    int c = (int)cc;

    if ((x < 0) || (x >=WIDTH))
        return;

    if (z<=0)
        return;

    if (z > ZBuffer[y][x]) {
        plot(x,y,c);
        ZBuffer[y][x] = z;
    }
}

void inline ScanAdd(FLOAT yy, FLOAT xx, FLOAT cc, FLOAT z)
{
    LONG x,y,c;

    x=(LONG)xx;
    y=(LONG)yy;
    c=(LONG)cc;
    if (y<0 || y>=HEIGHT)
        return;

    if (scanlines[y].Active) {
        if( (scanlines[y].Right.On == TRUE)) {
            if (x < scanlines[y].Left.x) {
                scanlines[y].Left.x  = x;
                scanlines[y].Left.z  = z;
                scanlines[y].Left.c  = c;
                scanlines[y].Left.On = TRUE;
            }
            else if (x > scanlines[y].Right.x) {
                scanlines[y].Right.x  = x;
                scanlines[y].Right.z  = z;
                scanlines[y].Right.c  = c;
                scanlines[y].Right.On = TRUE;
            }
        }
        else {
            if (x < scanlines[y].Left.x) {
                scanlines[y].Right = scanlines[y].Left;
                scanlines[y].Left.x  = x;
                scanlines[y].Left.z  = z;
                scanlines[y].Left.c  = c;
                scanlines[y].Left.On = TRUE;
            }
            else {
                scanlines[y].Right.x  = x;
                scanlines[y].Right.z  = z;
                scanlines[y].Right.c  = c;
                scanlines[y].Right.On = TRUE;
            }
        }
    }
    else {
        scanlines[y].Active = TRUE;
        scanlines[y].Left.x  = x;
        scanlines[y].Left.z  = z;
        scanlines[y].Left.c  = c;
        scanlines[y].Left.On = TRUE;
    }

}

void ZBufferTri(FPOLY *data)
{
    FPOINT *p1,*p2,*p3;
    FPOINT *f,*m,*l;
    LONG i;

    for(i=0; i<HEIGHT; i++) {
        scanlines[i].Active = FALSE;
        scanlines[i].Left.On = FALSE;
        scanlines[i].Right.On = FALSE;
    }

    p1=data->P1; p2=data->P2; p3=data->P3;

    //sort points
    if(p1->Y < p2->Y) {
        if(p1->Y < p3->Y) {
            if(p2->Y < p3->Y) {f=p1; m=p2; l=p3;}    //123
            else              {f=p1; m=p3; l=p2;}    //132
        } else                {f=p3; m=p1; l=p2;}    //312
    } else {
        if(p2->Y < p3->Y) {
            if(p1->Y < p3->Y) {f=p2; m=p1; l=p3;}    //213
            else              {f=p2; m=p3; l=p1;}    //231
        }else                 {f=p3; m=p2; l=p1;}    //321
    }

    //master y-clip
    if (
        f->Y > HEIGHT-1 ||
        l->Y < 0
    ) return;

    if (((f->X<0)&&(m->X<0)&&(l->X<0)) ||
        ((f->X>=WIDTH)&&(m->X>=WIDTH)&&(l->X>=WIDTH)))
        return;

    if (
        (f->X > 32767) || (f->X < -32768) ||
        (m->X > 32767) || (m->X < -32768) ||
        (l->X > 32767) || (l->X < -32768)
    ) return;

    if (f->Y == m->Y) {
        if (f->Y>=0) {
            ScanAdd(f->Y, f->X, f->Color, f->Z);
            ScanAdd(f->Y, m->X, m->Color, m->Z);
        }
    }
    else {
        INIT(xfm, f->X,     m->X,     (LONG)m->Y - (LONG)f->Y);
        INIT(zfm, f->Z,     m->Z,     (LONG)m->Y - (LONG)f->Y);
        INIT(cfm, f->Color, m->Color, (LONG)m->Y - (LONG)f->Y);

        for(i=(LONG)f->Y; i<m->Y; i++) {
            ScanAdd(i, VALUE(xfm),VALUE(cfm),VALUE(zfm));
            NEXT(xfm);
            NEXT(zfm);
            NEXT(cfm);
        }
    }


    if (m->Y == l->Y) {
        if (m->Y>=0) {
            ScanAdd(m->Y, m->X, m->Color, m->Z);
            ScanAdd(m->Y, l->X, l->Color, l->Z);
        }
    }
    else {
        INIT(xml, m->X,     l->X,     (LONG)l->Y - (LONG)m->Y);
        INIT(zml, m->Z,     l->Z,     (LONG)l->Y - (LONG)m->Y);
        INIT(cml, m->Color, l->Color, (LONG)l->Y - (LONG)m->Y);

        for(i=(LONG)m->Y; i<l->Y; i++) {
            ScanAdd(i, VALUE(xml),VALUE(cml),VALUE(zml));
            NEXT(xml);
            NEXT(zml);
            NEXT(cml);
        }
    }

    if (f->Y == l->Y) {
        if (f->Y>=0) {
            ScanAdd(f->Y, f->X, f->Color, f->Z);
            ScanAdd(f->Y, l->X, l->Color, l->Z);
        }
    }
    else {
        INIT(xfl, f->X,     l->X,     (LONG)l->Y - (LONG)f->Y);
        INIT(zfl, f->Z,     l->Z,     (LONG)l->Y - (LONG)f->Y);
        INIT(cfl, f->Color, l->Color, (LONG)l->Y - (LONG)f->Y);

        for(i=(LONG)f->Y; i<l->Y; i++) {
            ScanAdd(i, VALUE(xfl),VALUE(cfl),VALUE(zfl));
            NEXT(xfl);
            NEXT(zfl);
            NEXT(cfl);
        }
    }

    for(i=0; i<HEIGHT; i++) {
        if (scanlines[i].Active) {
            if (!scanlines[i].Right.On) {
                plotZPoint(scanlines[i].Left.x, i, scanlines[i].Left.z, scanlines[i].Left.c);
            }
            else {
#ifdef LOG
                fprintf(log, "From x:%d to x:%d\n",
                     scanlines[i].Left.x,
                     scanlines[i].Right.x);
#endif

                INIT(clr, scanlines[i].Left.c, scanlines[i].Right.c,
                          (LONG)scanlines[i].Right.x - (LONG)scanlines[i].Left.x);
                INIT(zlr, scanlines[i].Left.z, scanlines[i].Right.z,
                          (LONG)scanlines[i].Right.x - (LONG)scanlines[i].Left.x);
                for(int x=(int)scanlines[i].Left.x; x<=scanlines[i].Right.x; x++) {
                    plotZPoint(x, i, VALUE(zlr), VALUE(clr));
                    NEXT(clr);
                    NEXT(zlr);
                }
            }
        }
    }
}

#if 0
main()
{
    log = fopen("log.txt", "w");

    memset(ZBuffer, -1, sizeof(ZBuffer));

    init256(); pal();

    FPOINT p1, p2, p3;
    FPOLY p;

    p1.X = 0; p1.Y=0; p1.Z = 100; p1.Color = 255;
    p2.X = 300; p2.Y=100; p2.Z = 200; p2.Color = 128;
    p3.X = 10; p3.Y=180; p3.Z = 50; p3.Color = 0;

    p.P1 = &p1; p.P2 = &p2; p.P3 = &p3;
    ZBufferTri(&p);

    p1.X = 300; p1.Y=50; p1.Z = 50; p1.Color = 0;
    p2.X = 100; p2.Y=190; p2.Z = 10; p2.Color = 255;
    p3.X = 10; p3.Y=0; p3.Z = 200; p3.Color = 128;

    ZBufferTri(&p);

    waitkey();
    close256();

    fclose(log);
}
#endif
